"""Functional versions of torch.distributions."""
from torchsilk.distributions.base import Distribution as Distribution
from torchsilk.distributions.base import DistributionParams as DistributionParams
from torchsilk.distributions.base import DistributionParamsT as DistributionParamsT
from torchsilk.distributions.normal import Normal as Normal
from torchsilk.distributions.normal import NormalDistParams as NormalDistParams
